@@ -29,6 +29,14 @@
 from __future__ import absolute_import, division, print_function
 __metaclass__ = type
 
+import os
+import sys
+
+try:
+    import ansible_mitogen
+except ImportError:
+    sys.path.insert(0, os.path.abspath(os.path.join(__file__, '../../../..')))
+
 import ansible_mitogen.connection
 
 
